# ============================================================================
#    F A T 1 2   D i s k   C r e a t o r
#    Written by Tim Sarbin (tim.sarbin@essialsolutions.net)
# ============================================================================
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ============================================================================

from array import *
from struct import *
import os
import time

class Fat12Builder:
    def __init__(self, filename, bootfile):
        self.ATTR_READ_ONLY = 0x01
        self.ATTR_HIDDEN = 0x02
        self.ATTR_SYSTEM = 0x04
        self.ATTR_VOLUME_ID = 0x08
        self.ATTR_DIRECTORY = 0x10
        self.ATTR_ARCHIVE = 0x20
        # NOTE: These default values assume a standard 1.44mb floppy (emulated on a CD or actual)
        self.JmpBootCode = [0xEB, 0x1C, 0x90]
        self.OEMName = "JFOS"
        self.BytesPerSector = 512
        self.SectorsPerCluster = 1
        self.ReservedSectorCount = 1
        self.NumberOfFATs = 2
        self.RootEntriesCount = 224
        self.TotalSectors16 = 2880
        self.MediaType = 0xF0
        self.SectorsOccupiedByOneFAT = 9
        self.SectorsPerTrack = 18
        self.NumberOfHeads = 2
        self.HiddenSectors = 0
        self.TotalSectors32 = 0
        self.DriveNumber = 0 # Unused in JFOS
        self.VolumeID = 0xB007DA7A # BootData :)
        self.BootLabel = "JFOSBOOTDSK"
        self.FileSystemType = "FAT12   "
            
        self.DiskFile = open(filename, 'w+b')
        self.DiskFile.write(pack('BBB', self.JmpBootCode[0], self.JmpBootCode[1], self.JmpBootCode[2]))
        for i in range(1, 9):
            if (len(self.OEMName) < i):
                self.DiskFile.write(pack('c', " "))
            else:
                self.DiskFile.write(pack('c', self.OEMName[i-1]))
        self.DiskFile.write(pack('<H', self.BytesPerSector))
        self.DiskFile.write(pack('B', self.SectorsPerCluster))
        self.DiskFile.write(pack('<H', self.ReservedSectorCount))
        self.DiskFile.write(pack('B', self.NumberOfFATs))
        self.DiskFile.write(pack('<H', self.RootEntriesCount))
        self.DiskFile.write(pack('<H', self.TotalSectors16))
        self.DiskFile.write(pack('B', self.MediaType))
        self.DiskFile.write(pack('<H', self.SectorsOccupiedByOneFAT))
        self.DiskFile.write(pack('<H', self.SectorsPerTrack))
        self.DiskFile.write(pack('<H', self.NumberOfHeads))
        self.DiskFile.write(pack('<L', self.HiddenSectors))
        self.DiskFile.write(pack('<L', self.TotalSectors32))
        self.DiskFile.write(pack('B', self.DriveNumber))
        self.DiskFile.write(pack('B', 0))    # Reserved for use on WindowsNT
        self.DiskFile.write(pack('B', 0x29)) # Extended Boot Signature
        self.DiskFile.write(pack('<L', self.VolumeID))
        for i in range(1, 12):
            if (len(self.BootLabel) < i):
                self.DiskFile.write(pack('c', " "))
            else:
                self.DiskFile.write(pack('c', self.BootLabel[i-1]))
        for i in range(1, 9):
            if (len(self.FileSystemType) < i):
                self.DiskFile.write(pack('c', " "))
            else:
                self.DiskFile.write(pack('c', self.FileSystemType[i-1]))
        if (bootfile != ""):
            BootSecData = open(bootfile, "rb")
            for i in range(0,448):
                ByteData = BootSecData.read(1)
                if (len(ByteData) == 0):
                    self.DiskFile.write(pack('B', 0))
                else:
                    self.DiskFile.write(pack('B', ByteData[0]))
            BootSecData.close()
            self.DiskFile.write(pack('<H', 0xAA55))
        
        self.DiskFile.seek(1475072 - self.DiskFile.tell() - 1)
        self.DiskFile.write(pack('B', 0))
        
        self.WriteClusterEntry(0, 0xF00 + self.MediaType)
        self.WriteClusterEntry(1, 0xFFF)
        
    def CloseDisk(self):
        self.DiskFile.close()
        
    def MakeDOSFileName(self, filename):
        NewFileName = os.path.split(filename)[1]
        FATFilePartA = os.path.splitext(NewFileName)[0]
        FATFilePartB = os.path.splitext(NewFileName)[1][1:4]
        if ((len(FATFilePartA) > 8) or (len(FATFilePartA) < 1) or (len(FATFilePartB) > 3)):
            return ""
        FATFileFinal = FATFilePartA
        while (len(FATFileFinal) < 8):
            FATFileFinal += " "
        FATFileFinal += FATFilePartB
        return FATFileFinal.upper()

    
    def AddFile(self, filename, attributes):
        FileBeingAdded = open(filename, "rb")
        NewFileData = FileBeingAdded.read()
        FileBeingAdded.close()
        NewFileSize = os.path.getsize(filename)
        FATFileFinal = self.MakeDOSFileName(filename)
        if (FATFileFinal == ""):
            print("Error, '" + filename + "' cannot be formatted as a valid FAT12 file name.")
            return 0

        RootDirPos = ((self.ReservedSectorCount + (self.SectorsOccupiedByOneFAT * self.NumberOfFATs))  * self.BytesPerSector)

        for rootIdx in range(0, self.RootEntriesCount):
            self.DiskFile.seek(RootDirPos + (rootIdx * 32), 0)
            if ((self.DiskFile.peek(1)[0] == 0) or (self.DiskFile.peek(1)[0] == 0xE5)):
                break
            
        for i in range(1, 12):
            self.DiskFile.write(pack('c', FATFileFinal[i-1]))
        self.DiskFile.write(pack('B', attributes))
        self.DiskFile.write(pack('B', 0))
        
        # TODO: Date and TIME fields######################
        self.DiskFile.write(pack('B', 0))
        self.DiskFile.write(pack('<H', 0x0000))
        self.DiskFile.write(pack('<H', 0x0000))
        self.DiskFile.write(pack('<H', 0x0000))
        # #################################################
        
        self.DiskFile.write(pack('<H', 0x0000))
        
        # TODO: Date and TIME fields######################
        self.DiskFile.write(pack('<H', 0x0000))
        self.DiskFile.write(pack('<H', 0x0000))
        # #################################################

        CurClusterNum = self.AllocateCluster(0x00)
        ClusterStart = CurClusterNum
       
        self.DiskFile.seek(RootDirPos + (rootIdx * 32) + 26, 0)
        self.DiskFile.write(pack('<H', ClusterStart))
        self.DiskFile.write(pack('<L', len(NewFileData)))
        
        
        NumClusters = int(len(NewFileData) / self.GetClusterSizeInBytes())
        if (NumClusters != (len(NewFileData) / self.GetClusterSizeInBytes())):
            NumClusters += 1

       
        FileOffset = 0
        for i in range(0, NumClusters):
            if (i != 0):
                CurClusterNum = self.AllocateCluster(CurClusterNum)
            self.DiskFile.seek(self.GetClusterOffset(CurClusterNum), 0)
            self.DiskFile.write(NewFileData[FileOffset:FileOffset+self.GetClusterSizeInBytes()])
            FileOffset += self.GetClusterSizeInBytes()
        print("Added file: [" + FATFileFinal + "] " + str(NumClusters) + " cluster(s) starting at cluster " + str(ClusterStart))
        
    def DosCharsToStr(self, doschars):
        res = ""
        for i in range(0, len(doschars)):
            if (doschars[i] == 0x00):
                return res
            res += chr(doschars[i])
        return res
        
    def DeleteFile(self, filename):
        DosFileName = self.MakeDOSFileName(filename)
        if (DosFileName == ""):
            return 0
        RootDirPos = ((self.ReservedSectorCount + (self.SectorsOccupiedByOneFAT * self.NumberOfFATs))  * self.BytesPerSector)
        FileFound = 0
        filestruct = Struct("ccccccccccc")
        for rootIdx in range(0, self.RootEntriesCount):
            self.DiskFile.seek(RootDirPos + (rootIdx * 32), 0)
            CurFileName = self.DosCharsToStr(self.DiskFile.read(11))
            if (CurFileName == DosFileName):
                FileFound = 1
                break
        if (FileFound == 0):
            print("File '" + DosFileName + "' not found.")
            return 0
            
        self.DiskFile.seek(RootDirPos + (rootIdx * 32), 0)
        self.DiskFile.write(pack('B', 0xE5))
        self.DiskFile.seek(RootDirPos + (rootIdx * 32) + 26, 0)
        CurCluster = ord(self.DiskFile.read(1)) + (ord(self.DiskFile.read(1)) << 8)
        while (1):
            NextCluster = self.ReadClusterEntry(CurCluster)
            self.WriteClusterEntry(CurCluster, 0x000)
            if (NextCluster == 0xFFF):
                break
            CurCluster = NextCluster
        print("Deleted file: [" + DosFileName + "] ")
        return 1
        
    def WriteClusterEntry(self, clusterstart, nextcluster):
        for i in range(0,self.NumberOfFATs):
            FATPos = ((self.ReservedSectorCount + (self.SectorsOccupiedByOneFAT * i)) * self.BytesPerSector)
            ClusterBytePos = int(FATPos + (clusterstart * 1.5))
            self.DiskFile.seek(ClusterBytePos, 0)
            if (ClusterBytePos == (FATPos + (clusterstart * 1.5))):
                # We are at sector A in [AAAA AAAA][BBBB AAAA][BBBB BBBB]
                self.DiskFile.write(pack('B', nextcluster & 0xFF))
                NextVal = self.DiskFile.peek(1)[0]
                self.DiskFile.write(pack('B', ((nextcluster >> 8) + (NextVal & 0xF0))))
            else:
                # We are at sector B in [AAAA AAAA][BBBB AAAA][BBBB BBBB]
                PrevVal = self.DiskFile.peek(1)[0]
                self.DiskFile.write(pack('B', ((nextcluster << 4) & 0xF0 ) + (PrevVal & 0x0F)))
                self.DiskFile.write(pack('B', nextcluster >> 4))
            
    def ReadClusterEntry(self, cluster):
        FATPos = self.ReservedSectorCount * self.BytesPerSector
        ClusterBytePos = int(FATPos + (cluster * 1.5))
        self.DiskFile.seek(ClusterBytePos, 0)
        if (ClusterBytePos == (FATPos + (cluster * 1.5))):
            # We are at sector A in [AAAA AAAA][BBBB AAAA][BBBB BBBB]
            return ( ord(self.DiskFile.read(1)) + ((ord(self.DiskFile.read(1)) & 0x0F) << 8) )
        else:
            # We are at sector B in [AAAA AAAA][BBBB AAAA][BBBB BBBB]
            return ( ((ord(self.DiskFile.read(1)) & 0xF0) >> 4) + (ord(self.DiskFile.read(1)) << 4) )


    
    def IsLastCluster(self, cluster):
        if (self.ReadClusterEntry(cluster) != 0xFFF):
            return 0
        return 1
        
    def IsClusterInUse(self, cluster):
        if (self.ReadClusterEntry(cluster) == 0x00):
            return 0
        return 1
        
    def AllocateCluster(self, previouscluster):
        for i in range(0, int((self.SectorsOccupiedByOneFAT * self.BytesPerSector) / 2)):
            if (self.IsClusterInUse(i) == 0):
                break
        self.WriteClusterEntry(i, 0xFFF)
        if (previouscluster > 0):
            self.WriteClusterEntry(previouscluster, i)
        return i
        
    def GetClusterSizeInBytes(self):
        return self.BytesPerSector * self.SectorsPerCluster

    def GetClusterOffset(self, clusternum):
        return ((self.ReservedSectorCount + (self.SectorsOccupiedByOneFAT * self.NumberOfFATs)) * self.BytesPerSector) + (self.RootEntriesCount * 32) + ((clusternum - 2) * (self.SectorsPerCluster * self.BytesPerSector))